/* ------------------------------------------------------------------------------------
 * Un substitut de la commande Unix ls -laR
 * Les fichiers sont affichs dans l'ordre o ils se trouvent dans le 
 * rpertoire alors que le vrai ls les trie alphabtiquement.
 * Jean-Paul Rigault 
 * ------------------------------------------------------------------------------------
 */
#include <unistd.h>
#include <stdio.h>
#include <string.h>
#include <pwd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

/* Pour avoir PATH_MAX (on pourrait aussi utiliser pathconf() pour ne pas tre
 * li  Linux) */
#include <linux/limits.h>

/* Fonctions utilitaires de manipulation de noms de fichiers */
#include "util.h"

/* Imprimer le type du fichier et les droits d'accs
 * ------------------------------------------------- 
 *
 * On ne considre ici que deux types de fichiers : ordinaire et
 * rpertoire. Les autres sont indiqus par un '?'.
 */
void print_mode(mode_t mode) 
{
    char buff[11];
    char *p = buff;

    strcpy(buff, "drwxrwxrwx");

    if (S_ISREG(mode)) *p = '-';    	/* enlever le 'd' si ncessaire */
    else if (!S_ISDIR(mode)) *p = '?';

    for (++p; *p != '\0'; ++p, mode <<= 1)
        if (!(mode & S_IRUSR)) *p = '-'; /* enlever 'r' 'w' ou 'x' suivant le cas */

    printf(buff);
}

/* Formattage d'une date grce a strftime() */
char *format_date(time_t t)
{
    struct tm *date;
    static char buffer[30];

    date = localtime(&t); 
    strftime(buffer, 30, "%d/%m/%Y %H:%M", date);
    return buffer;
}

/* Impression de la ligne correspondant  une entre
 * -------------------------------------------------
 */
void print_fileinfo(char *filename)
{
    struct stat st;
    struct passwd *pass;
    char *owner;

    /* Recherche des attributs du fichier */
    if (stat(filename, &st) == -1)
    {
        fprintf(stderr, "stat() impossible sur %s\n", filename);
        return;
    }

    /* Recherche du nom du propritaire */
    pass = getpwuid(st.st_uid); 
    owner = pass ? pass->pw_name: "INCONNU";
  

    /* Impression des protections */
    print_mode(st.st_mode);

    /* Impression des infos complmentaires */ 
    printf(" %3d %-8s %7ld %s %s\n", 
           st.st_nlink, 		/* Nombre de liens */
           owner,			/* Nom en clair */
           st.st_size,			/* taille en octets */
           format_date(st.st_mtime),	/* date de modif */
           get_basename(filename));	/* Nom de base du fichier */
} 

/* Lister les infos d'un fichier ou d'un rpertoire
 * ------------------------------------------------
 */
void list(char *filename)
{ 
    DIR *dirp;
    struct dirent *dentry;
    char name[PATH_MAX + 1];
  
    if (is_dir(filename))
    {
        /* Nous avons affaire  un rpertoire */

        if ((dirp = opendir(filename)) == NULL) 
            fprintf(stderr, "rpertoire %s illisible\n", filename);
        else
        {
            /* On imprime le nom du rpertoire visit */
            printf("%s:\n", filename);

            /* 1ere phase: lister tout le monde */
            while ((dentry = readdir(dirp)) != NULL)
            {
                sprintf(name, "%s/%s", filename, dentry->d_name);
                print_fileinfo(name);
            }
      
            /* 2nde phase: appel rcursif pour les sous-rpertoires de filename */
            rewinddir(dirp);
            while ((dentry = readdir(dirp)) != NULL)
            {
                /* On vite "." et ".." (sinon boucle infinie ! */
                if (!is_dot_dir(dentry->d_name))
                {
                    sprintf(name, "%s/%s", filename, dentry->d_name);
                    if (is_dir(name)) list(name);
                }
            }
            closedir(dirp);
        }
    }
    else print_fileinfo(filename);
}

/* Programme principal
 * -------------------
 *
 * Invocation : lsrec.exe [file ...]
 * 
 * Sans argument on liste le rpertoire courant, sinon on liste (rcursivement)
 * tous les fichiers donns en argument
 */
int main(int argc, char **argv)
{
    if (argc == 1) list(".");
    else 
    {
         while (*++argv) list(*argv);
    }
    return 0;
}

