/* ------------------------------------------------------------------------------------
 * Copie de fichiers : premire version (simplifie)
 * Jean-Paul Rigault 
 * ------------------------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <sys/resource.h>

/* Pour avoir PATH_MAX (on pourrait aussi utiliser pathconf() pour ne pas tre
 * li  Linux) */
#include <linux/limits.h>

#include "util.h"

/* Buffer d'E/S */
#define DEFBUFFSIZE 1024		/* Taille par dfaut du buffer de copie */
static char *buffer;			/* Buffer de copie */
size_t buffer_size = DEFBUFFSIZE;	/* Taille effective du buffer de copie */

/* Options */
int option_b = 0;			/* Spcification de la taille de buffer */

/* Permissions par dfaut des fichiers crs : rw pour tous
 * (elles seront en plus filtres par le umask)
 */
#define USER_PERM (S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)

/* Erreur d'usage
 * --------------
 */
static void usage()
{
    fprintf(stderr, "Usage:\tmycp1 [-b buffsize] f1 f2\n"
                    "\tmycp1 [-b buffsize] f1 ... fn d2\n");
    exit(1);
}

/* Copie d'un fichier ordinaire dans un autre 
 * ------------------------------------------
*/
static int nbchars = 0; 	/* Nombre total de caractres transfrs */

void copy(const char *src, const char *dest)
{
    int fdsrc;
    int fddest;
    int n;
 
    /* Ouvrir src en lecture seule */
    if ((fdsrc = open(src, O_RDONLY)) < 0)
    {
        perror("mycp1 : fichier origine");
        fprintf(stderr, "\tImpossible d'ouvrir fichier origine %s\n", src);
        exit(1);
    }

    /* Ouvrir dest en criture, le crer s'il n'existe par, le tronquer  0
     * s'il existe dj. */
    if ((fddest = open(dest, O_WRONLY|O_CREAT|O_TRUNC, USER_PERM)) < 0)
    {
        perror("mycp1 : fichier destination");
        fprintf(stderr, "\tImpossible d'ouvrir/crer fichier destination %s\n", dest);
        exit(1);
    }

    /* Les fichiers sont ouverts; copier */
    while ((n = read(fdsrc, buffer, buffer_size)) > 0) 
    {
        if (n < 0)
        {
            perror("mycp1 : lecture");
            fprintf(stderr, "\tErreur de lecture fichier %s\n", src);
            exit(1);
        }
        if (write(fddest, buffer, n) < n)
        {
            perror("mycp1 : CRITURE");
            fprintf(stderr, "\tErreur d'criture fichier %s\n", dest);
            exit(1);
        }
        nbchars += n;
    }
  
    /* Fermeture des fichiers */
    close(fdsrc); 
    close(fddest);
}

/* Copie des fichiers
 * ------------------
*/
void copy_files(const char *src, const char *dest)
{
    char newfile[PATH_MAX];

    /* Dans cette version, on ne copie pas les rpertoires */
    if (is_dir(src))
    {
        fprintf(stderr, "mycp1: %s est un rpertoire (non copi)\n", src);
        exit(1);
    }

    /* Si la destination est un rpertoire, on cre un nouveau fichier ou on
     * crase un fichier dj existant dans ce rpertoire */
    if (is_dir(dest))
    {
        sprintf(newfile, "%s/%s", dest, get_basename(src));
        dest = newfile;
    }

    /* On vrifie qu'on ne copie pas un fichier sur lui-mme (ceci peut se
     * produire  cause des *liens* */
    if (same_file(src, dest))
    {
        fprintf(stderr, "mycp1: %s et %s sont identiques (non copi)\n", src, dest);
        exit(1);
    }

    /* Copie */
    copy(src, dest);
}

/* Convertir une struct timeval en un nombre rel de secondes */
#define TIMEVAL(tv) ((double)(tv).tv_sec + 1.0e-06 * (double)(tv).tv_usec)

/* Imprime quelques statistiques de performances */
static void print_rusage(const struct rusage *ru0, const struct rusage *ru1)
{
    double delta_t;

    delta_t = TIMEVAL(ru1->ru_utime) - TIMEVAL(ru0->ru_utime);
    printf("Temps utilisateur : %.3f : %.2f kcar/s\n",
           delta_t, (double)nbchars / delta_t / (1024 * 1024));

    delta_t = TIMEVAL(ru1->ru_stime) - TIMEVAL(ru0->ru_stime);
    printf("Temps systme : %.3f  : %.2f kcar/s\n",
           delta_t, (double)nbchars / delta_t / (1024 * 1024));
}

/* Programme principal
 * -------------------
 *
 * Invocation : cp [-b buffsize] filesrc filedest
 *              cp [-b buffsize] filesrc1 [filesrc2 ...]  filedest
 *
 * filedest peut tre un rpertoire.
 */
int main(int argc, char *argv[])
{
    int ifiles = 1;	/* index du dbut des arguments correspondant aux
                         * fichiers  copier */
    char *dest;		/* destination de la copie */
    int n; 		/* nombre de fichiers  copier */
    int i;
    struct rusage ru0;	/* utilisation des ressources systme */
    struct rusage ru1;	/* utilisation des ressources systme */

    /* Dcodage de l'argument '-b buffsize' ventuel */
    if (argv[1][0] == '-') 
    {
        if (argv[1][1] == 'b' && argv[1][2] == '\0' && argc >= 2)
        {
            buffer_size = atoi(argv[2]);
            if (buffer_size <= 0) usage();
            ifiles = 3; 
            ++option_b;
        }
        else usage();
    }

    /* Recherche de la destination (le dernier paramtre) */    
    dest = argv[--argc];

    /* Nombre de fichiers  copier */
    n = argc - ifiles;
    
    /* On vrifie qu'il y a des fichiers  copier */
    if (n <= 0) usage();

    /* Si il y a plus de 1 fichier  copier, la destination doit tre un rpertoire */
    if (n > 1 && !is_dir(dest)) usage();

    /* Allocation du buffer de copie */
    buffer = malloc(buffer_size);

    /* Excution de la copie */
    getrusage(RUSAGE_SELF, &ru0);
    for (i = ifiles; i < argc; ++i) copy_files(argv[i], dest);
    getrusage(RUSAGE_SELF, &ru1);    
    if (option_b) print_rusage(&ru0, &ru1);

    return 0;
}
