/* ------------------------------------------------------------------------------------
 * Copie de fichiers : deuxime version (options -v, -r, -i)
 * Jean-Paul Rigault 
 * ------------------------------------------------------------------------------------
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <dirent.h>

/* Pour avoir PATH_MAX (on pourrait aussi utiliser pathconf() pour ne pas tre
 * li  Linux) */
#include <linux/limits.h>

#include "util.h"

/* Buffer d'E/S */
#define DEFBUFFSIZE 1024		/* Taille par dfaut du buffer de copie */
static char *buffer;			/* Buffer de copie */
size_t buffer_size = DEFBUFFSIZE;	/* Taille effective du buffer de copie */

/* Permissions par dfaut des fichiers crs : rw pour tous
 * (elles seront en plus filtres par le umask)
 */
#define USER_PERM (S_IRUSR|S_IWUSR|S_IRGRP|S_IWGRP|S_IROTH|S_IWOTH)
#define DIR_PERM (USER_PERM|S_IXUSR|S_IXGRP|S_IXOTH)

/* Options de la ligne de commande */
int option_v = 0;	/* copie verbeuse */
int option_r = 0;	/* copie rcursive */
int option_i = 0;	/* copie avec demande de confirmation d'crasement de fichier */

int multiple_files = 0; /* plusieurs fichiers  copier */

/* Erreur d'usage
 * --------------
 *
 * On ne revient pas de cette fonction.
 */
static void usage()
{
    fprintf(stderr, "Usage:\tmycp2 [options] f1 f2\n"
                    "\tmycp2 [options] f1 ... fn d2\n");
    fprintf(stderr, "\no les options sont :\n"
                    "\t-b buffsize\ttaille du buffer d'E/S\n"
                    "\t-v\tcopie verbeuse\n"
                    "\t-i\tdemande de confirmation si destination existe\n"
                    "\t-r\tcopie rcursive des rpertoires\n");
    exit(1);
}

/* Copie d'un fichier ordinaire dans un autre 
 * ------------------------------------------
*/
void copy(char *src, char *dest)
{
    int fdsrc;
    int fddest;
    int n;
 
    /* Ouvrir src en lecture seule */
    if ((fdsrc = open(src, O_RDONLY)) < 0)
    {
        perror("mycp2 : fichier origine");
        fprintf(stderr, "\tImpossible d'ouvrir fichier origine %s\n", src);
        exit(1);
    }

    /* Ouvrir dest en criture, le crer s'il n'existe par, le tronquer  0
     * s'il existe dj. */
    if ((fddest = open(dest, O_WRONLY|O_CREAT|O_TRUNC, USER_PERM)) < 0)
    {
        perror("mycp2 : fichier destination");
        fprintf(stderr, "\tImpossible d'ouvrir/crer fichier destination %s\n", dest);
        exit(1);
    }

    if (option_v) printf("mycp2 : copie de %s sur %s\n", src, dest);

    /* Les fichiers sont ouverts; copier */
    while ((n = read(fdsrc, buffer, buffer_size)) > 0) 
    {
        if (n < 0) 
        {
            perror("mycp2 : lecture");
            fprintf(stderr, "\tErreur de lecture fichier %s\n", src);
            exit(1);
        }
        if (write(fddest, buffer, n) < n) 
        {
            perror("mycp2 : CRITURE");
            fprintf(stderr, "\tErreur d'criture fichier %s\n", dest);
            exit(1);
        }
    }
  
    /* Fermeture des fichiers */
    close(fdsrc); 
    close(fddest);
}

/* Copie des fichiers
 * ------------------
*/
void copy_files(char *src, char *dest)
{
    char newsrc[PATH_MAX];
    char newdest[PATH_MAX];
    DIR *dirp;
    struct dirent *dentry;

    if (is_dir(src))
    {
        /* Il faut l'option -r pour dclencher la copie rcursive */
        if (!option_r)
        {
            fprintf(stderr, "mycp2: %s est un rpertoire (non copi)\n", src);
            exit(1);
        }

        /* La destination doit tre un rpertoire */
        if (file_exists(dest)) sprintf(newdest, "%s/%s", dest, get_basename(src));
        else strcpy(newdest, dest);
        mkdir(newdest, DIR_PERM);	/* Inutile de tester; il y aura une
                                         * erreur plus tard si le rpertoire
                                         * n'a pas pu tre cr */

        /* Copie rcursive */
        if ((dirp = opendir(src)) == NULL)
        {
            fprintf(stderr, "mycp2 : impossible ouvrir rpertoire %s\n", src);
            exit(1);
        }

        for (dentry = readdir(dirp); dentry != NULL; dentry = readdir(dirp))
        {
            if (is_dot_dir(dentry->d_name)) continue;
            sprintf(newsrc, "%s/%s", src, dentry->d_name);
            copy_files(newsrc, newdest);
        }

        closedir(dirp);
 
    }
    else 
    {
        /* Copie d'un fichier ordinaire */

        /* Si la destination n'existe pas et qu'on a plusieurs fichiers 
         * copier, il faut crer un rpertoire */
        if (multiple_files && !file_exists(dest))  mkdir(dest, DIR_PERM);

        /* Si la destination est un rpertoire, on cre un nouveau fichier ou on
         * crase un fichier dj existant dans ce rpertoire */
        if (is_dir(dest)) 
        {
            sprintf(newdest, "%s/%s", dest, get_basename(src));
            dest = newdest;
        }

        /* Si la destination existe et si on est interactif (-i) demander
         * confirmation */
        if (option_i && file_exists(dest)) 
        {   
            int rep;

            fprintf(stderr, "mycp2 : destination %s existe, l'craser ? ([yYoOnN]) ", dest); 
            fflush(stderr);
            rep = getchar();
            while (getchar() != '\n') {} /* oublie la fin de ligne */
            if (rep != 'y' && rep != 'Y' && rep != 'o' && rep != 'O') return;
        }

        /* On vrifie qu'on ne copie pas un fichier sur lui-mme (ceci peut se
         * produire  cause des *liens* */
        if (same_file(src, dest))
        {
            fprintf(stderr, "mycp2: %s et %s sont identiques (non copi)\n", src, dest);
            exit(1);
        }

        /* Copie */
        copy(src, dest);
    }
}

/* Programme principal
 * -------------------
 *
 * Invocation : cp [options] filesrc filedest
 *              cp [options] filesrc1 [filesrc2 ...]  filedest
 *
 * filedest peut tre un rpertoire.
 */
int main(int argc, char *argv[])
{
    int ifiles = 1;	/* index du dbut des arguments correspondant aux
                         * fichiers  copier */
    char *dest;		/* destination de la copie */
    int n; 		/* nombre de fichiers  copier */
    int i;

    /* Dcodage des options */
    for (i = 1; i < argc; ++i)
    {
        if (argv[i][0] == '-') 
        {
            if (argv[i][2] != '\0') usage();

            switch (argv[i][1])
            {
            case 'b':
                if (++i >= argc) usage();
                buffer_size = atoi(argv[i]);
                if (buffer_size <= 0) usage();
                ifiles += 2; 	
                break;
            case 'v':
                ++option_v;
                ++ifiles;
                break;
            case 'r':
                ++option_r;
                ++ifiles;
                break;
            case 'i':
                ++option_i;
                ++ifiles;
                break;
            default:
                usage();
                break;                
            }
        }
        else break; /* dbut de la liste des fichiers  copier */
    }

    /* Recherche de la destination (le dernier paramtre) */    
    dest = argv[--argc];

    /* Nombre de fichiers  copier */
    n = argc - ifiles;
    
    /* On vrifie qu'il y a des fichiers  copier */
    if (n <= 0) usage();

    /* */
    if (n > 1) 
    {
        if (!option_r) usage();
        ++multiple_files;
    }

    /* Allocation du buffer de copie */
    buffer = malloc(buffer_size);

    /* Excution de la copie */
    for (i = ifiles; i < argc; ++i) copy_files(argv[i], dest);

    return 0;
}
